"""BI CLI."""

import json
import os
import webbrowser
from datetime import UTC, datetime, timedelta
from enum import StrEnum, auto
from typing import Annotated, Optional

import typer
from rich.table import Table
from slices_bi_client.exceptions import BadRequestException, ForbiddenException, NotFoundException
from slices_bi_client.models.compute_resource_create import ComputeResourceCreate
from slices_bi_client.models.network_interface_create import NetworkInterfaceCreate
from slices_bi_client.models.new_resources_request import NewResourcesRequest
from slices_bi_client.models.reboot_action import RebootAction
from slices_bi_client.models.resource_action import ResourceAction
from slices_bi_client.models.resource_state import ResourceState
from slices_bi_client.models.ticket import Ticket
from slices_bi_client.models.update_resources_request import UpdateResourcesRequest
from slices_bi_client.models.vmlab_backend_schemas_bi_resource_reboot_action_action import (
    VmlabBackendSchemasBiResourceRebootActionAction,
)
from slices_expauth_client.models.experiment import Experiment
from slices_expauth_client.models.experiment_update import ExperimentUpdate

from . import console, err_console
from .bi_context import BIContext, BIContextObj, _is_type_id
from .bi_diskimage import disk_images_app
from .bi_flavor import flavors_app
from .context import Context
from .progress import follow_progress
from .util.ssh_known_hosts import SSHKnownHostsManager

bi_app = typer.Typer(short_help="Basic Infrastructure", no_args_is_help=True)
bi_app.add_typer(flavors_app, name="flavor", short_help="Flavors.")
bi_app.add_typer(disk_images_app, name="diskimage", short_help="Disk images.")


def bi_site_callback(ctx: Context, _param: typer.CallbackParam, value: str):
    """Validate bi_site value."""
    if value not in ctx.obj.api.config.bi:
        raise typer.BadParameter("Invalid BI Site ID")

    return value


@bi_app.callback(invoke_without_command=True)
def initialize_bi(
    ctx: Context,
    site_id: Annotated[str, typer.Option(envvar="SLICES_BI_SITE_ID", callback=bi_site_callback)] = "vmlab",
    cluster_id: Annotated[str, typer.Option(envvar="SLICES_BI_CLUSTER_ID")] = "default",
):
    """Initialize BI Context."""
    ctx.obj = BIContextObj(
        api=ctx.obj.api,
        user_info=ctx.obj.user_info,
        site_id=site_id,
        cluster_id=cluster_id,
    )


def _get_experiment(ctx: BIContext, experiment_id) -> Experiment:
    if _is_type_id(experiment_id, required_type="exp"):
        exp_obj = ctx.obj.api.experiments_api.get_experiment(experiment_id)
    else:
        exp_obj = ctx.obj.api.get_experiment(experiment_id)

    if not exp_obj:
        err_console.print(f"Could not retrieve experiment {experiment_id} :x:")
        raise typer.Exit(code=1)
    return exp_obj


def _get_registered_pubkeys(ctx: BIContext, experiment: Experiment) -> list[str]:
    """Get public keys of all members in the experiment."""

    experiment_members = ctx.obj.api.experiments_api.get_experiment_members(experiment.id)

    return [
        pubkey.public_key
        for member in experiment_members.experiment_members
        for pubkey in ctx.obj.api.pubkey_api.get_user_pubkeys(member.user_id).public_keys
    ]


def _find_flavor_id(ctx: BIContext, flavor_any_id: str) -> str:
    flavors_page = ctx.obj.bi_flavors_api.get_flavors()
    for flavor_inner in flavors_page.flavors:
        flavor = flavor_inner.actual_instance
        if flavor_any_id == flavor.id or flavor_any_id.lower() == flavor.friendly_name.lower():
            return flavor.id
    err_console.print(f"[red]Unknown flavor {flavor_any_id!r} :mag:")
    err_console.print("Check: [blue]slices bi flavor list")
    raise typer.Exit(code=1)


@bi_app.command()
def create(
    ctx: BIContext,
    name: Annotated[str, typer.Option(prompt=True)],
    image: Annotated[str, typer.Option(..., help="ID of the VM image", prompt=True)],
    experiment_id: Annotated[Optional[str], typer.Option()] = None,
    experiment_name: Annotated[Optional[str], typer.Option()] = None,
    flavor: Annotated[str, typer.Option(..., help="ID of the flavor")] = "xs",
    user_data: Annotated[Optional[typer.FileText], typer.Option()] = None,
    duration: Annotated[
        int, typer.Option(min=1, max=90 * 24, help="Lifetime of the resource in hours. Can be extended later on.")
    ] = 24,
    wait: Annotated[bool, typer.Option(help="Wait for resource to be created.")] = False,
    public_ipv4: Annotated[bool, typer.Option(help="Request a public IPv4.")] = False,
):
    """Request compute resource."""

    expires_at = datetime.now(UTC).replace(second=0, microsecond=0) + timedelta(hours=duration)

    if experiment_id:
        experiment = ctx.obj.api.experiments_api.get_experiment(experiment_id)
        project_id = experiment.project_id

    else:
        project_id = ctx.obj.get_current_project_id()
        if not project_id:
            err_console.print("[bold red]You must set a project first[/]")
            err_console.print("[red]Use: 'slices project use <project>'")
            raise typer.Exit(1)

        experiment = ctx.obj.api.get_or_create_experiment(
            experiment_name or typer.prompt("Experiment name"),
            project_id,
            expires_at=expires_at,
        )
    # if experiment expiration is shorter than requested duration, extend the experiment
    if expires_at > experiment.expires_at:
        ctx.obj.api.experiments_api.update_experiment(experiment.id, ExperimentUpdate(expires_at=expires_at))

    ssh_authorized_keys = _get_registered_pubkeys(ctx, experiment)
    if not ssh_authorized_keys:
        err_console.print("[red]You need to have at least one public SSH key registered. :x:")
        err_console.print("Use: [blue]slices pubkey register")
        raise typer.Exit(code=1)

    flavor_id = _find_flavor_id(ctx, flavor)

    nrr = NewResourcesRequest(
        expires_at=expires_at,
        resources=[
            ComputeResourceCreate(
                friendly_name=name,
                disk_image_id=image,
                flavor_id=flavor_id,
                userdata=user_data.read() if user_data else None,
                network_interfaces=[
                    NetworkInterfaceCreate(
                        port_id="eth0",
                        network_id="private" if not public_ipv4 else "public",
                        dhcp4=None,
                        dhcp6=None,
                        addresses=None,
                        nameservers=None,
                        routes=None,
                    )
                ],
            )
        ],
        ssh_authorized_keys=ssh_authorized_keys,
        tickets=[
            Ticket(
                jwt=ctx.obj.api.get_id_token_for_audience(ctx.obj.api.config.bi[ctx.obj.site_id].audience, project_id)
            )
        ],
    )

    try:
        response = ctx.obj.get_bi_resources_api_by_experiment(experiment).request_experiment_resources(
            project=project_id,
            experiment_id=experiment.id,
            new_resources_request=nrr,
        )
    except BadRequestException as exc:
        err_console.print("[bold red]Could not create resource[/] :x:")
        err_console.print("[red]" + json.loads(exc.body)["detail"])
        raise typer.Exit(code=1) from None
    except ForbiddenException as exc:
        err_console.print("[bold red]No permission to create resource[/] :x:")
        err_console.print("[red]" + json.loads(exc.body)["detail"])
        raise typer.Exit(code=1) from None

    if not len(response.resources):
        err_console.print("[bold red]No resources were created[/] :x:")

    if wait:
        follow_progress(
            ctx.obj.bi_tasks_api,
            response.resources[0].task_id,
            f"Create resource {name}",
            transient=True,
        )

    for resource in response.resources:
        console.print(f"[bold green]:sparkles: Created {resource.friendly_name} with ID {resource.id}")


@bi_app.command("list")
def list_experiment_resources(
    ctx: BIContext,
    experiment: Annotated[str, typer.Argument()],
):
    """List experiment resources."""
    exp = _get_experiment(ctx, experiment)

    experiment_resources = ctx.obj.get_bi_resources_api_by_experiment(exp).get_experiment_resources(
        experiment_id=exp.id,
        project=exp.project_id,
    )

    table = Table(
        "ID",
        "Name",
        "Status",
        "Created At",
        "Expires At",
        "Public IPv4",
        "Private IPv4",
        title=f"Resources in Experiment {exp.id}",
    )
    for r in experiment_resources.resources:
        table.add_row(
            r.id,
            r.friendly_name,
            r.status,
            r.created_at.astimezone().strftime("%Y-%m-%d %H:%M"),
            r.expires_at.astimezone().strftime("%Y-%m-%d %H:%M"),
            r.public_ipv4,
            r.private_ipv4,
        )

    console.print(table)


@bi_app.command("reset")
def reset_resource(
    ctx: BIContext,
    experiment: Annotated[str, typer.Argument()],
    resource: Annotated[str, typer.Argument],
):
    """Perform a hard reset a resource."""
    exp = _get_experiment(ctx, experiment)

    if not _is_type_id(resource, required_type="r"):
        err_console.print("A valid resource ID is required")
        raise typer.Exit(code=1)

    task = ctx.obj.get_bi_resources_api_by_experiment(exp).perform_project_resource_action(
        exp.project_id,
        resource,
        ResourceAction(RebootAction(reboot=VmlabBackendSchemasBiResourceRebootActionAction(type="HARD"))),
    )

    project_id = ctx.obj.get_current_project_id()
    if not project_id:
        raise typer.Exit(0)

    follow_progress(
        ctx.obj.bi_tasks_api,
        task.id,
        f"Resetting {resource}",
    )


@bi_app.command("destroy")
def destroy_resource(
    ctx: BIContext,
    experiment: Annotated[str, typer.Argument()],
    resource: Annotated[str, typer.Argument],
):
    """Destroy a resource."""
    exp = _get_experiment(ctx, experiment)

    if not _is_type_id(resource, required_type="r"):
        err_console.print("A valid resource ID is required")
        raise typer.Exit(code=1)

    try:
        task = ctx.obj.get_bi_resources_api_by_experiment(exp).delete_experiment_resource(
            exp.project_id, exp.id, resource
        )
    except NotFoundException:
        err_console.print(f"[red]{resource} was not found :x:")
    else:
        SSHKnownHostsManager().cleanup(resource)

        project_id = ctx.obj.get_current_project_id()
        if not project_id:
            raise typer.Exit(0)

        follow_progress(
            ctx.obj.bi_tasks_api,
            task.id,
            f"Deleting {resource}",
        )


@bi_app.command("renew")
def renew_resource(
    ctx: BIContext,
    experiment: Annotated[str, typer.Argument()],
    resource: Annotated[str, typer.Argument],
    duration: Annotated[
        int,
        typer.Option(min=1, max=90 * 24, help="New lifetime of the resource in hours."),
    ],
):
    """Extend the expiration date of a resource."""
    exp = _get_experiment(ctx, experiment)

    new_expire_at = datetime.now(UTC).replace(second=0, microsecond=0) + timedelta(hours=duration)

    # if experiment expiration is shorter than requested duration, extend the experiment
    if new_expire_at > exp.expires_at:
        err_console.print("Renewing Experiment...")
        ctx.obj.api.experiments_api.update_experiment(exp.id, ExperimentUpdate(expires_at=new_expire_at))

    if not _is_type_id(resource, required_type="r"):
        err_console.print("A valid resource ID is required")
        raise typer.Exit(code=1)

    try:
        err_console.print("Renewing Resource...")
        updated_resource = ctx.obj.get_bi_resources_api_by_experiment(exp).update_resource(
            exp.project_id, exp.id, resource, UpdateResourcesRequest(expires_at=new_expire_at)
        )
    except NotFoundException:
        err_console.print(f"[red]{resource} was not found :x:")
        raise typer.Exit(1) from None

    err_console.print(
        f"[green]:heavy_check_mark: Done.[/green] Resource now expires at "
        f"{updated_resource.expires_at.astimezone().strftime('%Y-%m-%d %H:%M')}."
    )
    project_id = ctx.obj.get_current_project_id()
    if not project_id:
        raise typer.Exit(0)


class ProxyOption(StrEnum):
    """Proxy Options."""

    off = auto()
    on = auto()
    auto = auto()


class ShowOption(StrEnum):
    """Show options."""

    nothing = auto()
    command = auto()
    ansible = auto()
    ssh_config = auto()


DEFAULT_SSH_PORT = 22


@bi_app.command("ssh")
def ssh_to_resource(  # noqa: C901
    ctx: BIContext,
    experiment: Annotated[str, typer.Argument()],
    resource: Annotated[str, typer.Argument],
    no_exec: Annotated[
        bool,
        typer.Option(help="Do not call SSH to connect to the resource. Implied if --show is ansible or ssh_config."),
    ] = False,
    show: Annotated[
        ShowOption,
        typer.Option(help="What to show: Nothing, SSH Command, Ansible inventory line, or lines for ~/.ssh/config"),
    ] = ShowOption.command,
    proxy: Annotated[ProxyOption, typer.Option(help="Use an SSH jump proxy.", case_sensitive=False)] = ProxyOption.auto,
):
    """Log in to a resource using SSH. Also show the login command."""
    exp = _get_experiment(ctx, experiment)

    if not _is_type_id(resource, required_type="r"):
        err_console.print("A valid resource ID is required")
        raise typer.Exit(code=1)

    compute_resource = ctx.obj.get_bi_resources_api_by_experiment(exp).get_experiment_resource(
        exp.project_id, exp.id, resource
    )
    if not compute_resource.ssh_logins:
        err_console.print("The resource has no SSH logins")
        raise typer.Exit(code=1)

    ssh_login = compute_resource.ssh_logins[0]

    destination = f"{ssh_login.username}@{ssh_login.host}"
    if ssh_login.port != DEFAULT_SSH_PORT:
        destination += f":{ssh_login.port}"

    if proxy == ProxyOption.auto:
        if compute_resource.public_ipv4 and compute_resource.public_ipv4 in ssh_login.host:
            proxy = ProxyOption.off
        else:
            # For now, always turn on.
            # TODO detect if VPN is on, and the private IP is reachable directly
            proxy = ProxyOption.on

    console.print(f"proxy={proxy} ssh_login.jump_proxy={ssh_login.jump_proxy}")

    ssh_command_options = []
    ansible_ssh_common_args = ""
    ssh_config_jump_proxy = ""
    if proxy == ProxyOption.on and ssh_login.jump_proxy:
        port_dest = f":{ssh_login.jump_proxy.port}" if ssh_login.jump_proxy.port != DEFAULT_SSH_PORT else ""
        ssh_jump_proxy_destination = f"{ssh_login.jump_proxy.username}@{ssh_login.jump_proxy.host}{port_dest}"
        ssh_command_options.extend(["-J", ssh_jump_proxy_destination])
        ansible_ssh_common_args = f" ansible_ssh_common_args='-J {ssh_jump_proxy_destination}'"
        ssh_config_jump_proxy = f"JumpProxy {ssh_jump_proxy_destination}"

    command = ["ssh", *ssh_command_options, destination]

    command_line = " ".join(command)

    do_exec_ssh = (not no_exec) and show in [ShowOption.command, ShowOption.nothing]

    if ssh_login.host_keys and compute_resource.status in (
        ResourceState.IMAGING,
        ResourceState.BOOTING,
        ResourceState.INITIALIZING,
        ResourceState.STARTING,
        ResourceState.UP,
    ):
        SSHKnownHostsManager().cleanup_expired()
        SSHKnownHostsManager().add_login(compute_resource, ssh_login)

    if show == ShowOption.command:
        if do_exec_ssh:
            console.print(f"[bold blue]{command_line}")
        else:
            typer.echo(command_line)  # output without any markup

    elif show == ShowOption.ansible:
        # TODO if ComputeResource ever has a user id for this host, use that
        # output without any markup
        typer.echo(
            f"{ssh_login.host} "
            f"ansible_ssh_port={ssh_login.port} "
            f"ansible_ssh_host={ssh_login.host} "
            f"ansible_ssh_user={ssh_login.username} {ansible_ssh_common_args}"
        )

    elif show == ShowOption.ssh_config:
        # output without any markup
        typer.echo(f"""Host={ssh_login.host}
   HostName {ssh_login.host}
   Port {ssh_login.port}
   User {ssh_login.username}
   PreferredAuthentications publickey
   IdentitiesOnly yes
   {ssh_config_jump_proxy}""")

    if do_exec_ssh:
        os.execvp(command[0], command)  # noqa: S606


@bi_app.command("console")
def console_to_resource(  # noqa: C901
    ctx: BIContext,
    experiment: Annotated[str, typer.Argument()],
    resource: Annotated[str, typer.Argument],
    show_only: Annotated[
        bool,
        typer.Option(help="Do not try to open console URL, only show it."),
    ] = False,
):
    """Open a browser with a console to a resource. Also print the link that is opened."""
    exp = _get_experiment(ctx, experiment)

    if not _is_type_id(resource, required_type="r"):
        err_console.print("A valid resource ID is required")
        raise typer.Exit(code=1)

    compute_resource = ctx.obj.get_bi_resources_api_by_experiment(exp).get_experiment_resource(
        exp.project_id, exp.id, resource
    )
    if not compute_resource.console_url:
        err_console.print("The resource has no remote console")
        raise typer.Exit(code=1)

    console.print(f"Console URL: [bold blue]{compute_resource.console_url}")

    if not show_only:
        webbrowser.open(compute_resource.console_url, new=1)
