"""Custom slices-cli context."""

from dataclasses import dataclass
from typing import Literal

import typer
from typer import Context as BaseContext

from . import err_console
from .api import ApiClients
from .config import get_config_string
from .oauth2.token import AuthJWT


@dataclass
class ContextObj:
    """Context Object."""

    api: ApiClients
    user_info: AuthJWT | None

    def get_logged_in_user(self) -> AuthJWT:
        """Get user info or show error."""
        if not self.user_info:
            err_console.print("[red]You must be logged in to perform this action. :key:")
            raise typer.Exit(128)
        return self.user_info

    def get_logged_in_userinfo(self, field_name: Literal["sub", "preferred_username"]):
        """Return logged in user info."""
        return self.get_logged_in_user()[field_name]

    def get_current_project_id(self) -> str | None:
        """Return the current project."""
        return get_config_string("default", section="project")

    def get_current_project_token(self) -> str | None:
        """Return token for current project."""
        project_id = self.get_current_project_id()
        if not project_id:
            return None

        return self.api.get_project_jwt(project_id)


class Context(BaseContext):
    """Custom context for slices-cli."""

    obj: ContextObj
