"""Experiment CLI."""

from datetime import UTC, datetime
from typing import Annotated, Optional

import typer
from rich.table import Table
from slices_expauth_client.exceptions import BadRequestException
from slices_expauth_client.models.experiment_create import ExperimentCreate

from . import console, err_console
from .context import Context
from .utils import datetime_with_local_timezone, friendly_datetime, get_error_detail, is_experiment_id

experiment_app = typer.Typer(short_help="Experiments", no_args_is_help=True)


@experiment_app.command(name="create")
def create_experiment(
    ctx: Context,
    name: Annotated[str, typer.Option()],
    expiration: Annotated[datetime, typer.Option()],
    description: Annotated[Optional[str], typer.Option()] = None,
    project_id: Annotated[Optional[str], typer.Option()] = None,
):
    """Create a new experiment."""

    if project_id:
        project_token = ctx.obj.api.get_project_jwt(project_id)
    else:
        project_token = ctx.obj.get_current_project_token()

        if not project_token:
            err_console.print(
                "[bold red]No default project set. "
                "Please specify a default project with `slices project use <name>`, or pass a project id. :x:"
            )
            raise typer.Exit(1)

    try:
        experiment = ctx.obj.api.experiments_api.create_experiment(
            ExperimentCreate(
                friendly_name=name,
                description=description,
                expires_at=datetime_with_local_timezone(expiration),
                project_token=project_token,
            )
        )

    except BadRequestException as err:
        err_console.print("[bold red]Could not create experiment :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err

    console.print(
        f"[green]:sparkles: Created experiment [bold]{experiment.friendly_name}[/bold] "
        f"that expires on [bold]{friendly_datetime(experiment.expires_at)}[/]",
    )
    console.print(f"Experiment ID: {experiment.id}")


@experiment_app.command(name="list")
def list_experiments(
    ctx: Context,
    include_expired: Annotated[
        Optional[bool], typer.Option("--include-expired", "-e", help="Also show expired experiments.")
    ] = False,
    include_archived: Annotated[
        Optional[bool], typer.Option("--include-archived", "-A", help="Also show archived experiments.")
    ] = False,
    all: Annotated[Optional[bool], typer.Option("--all", "-a", help="Show all experiments.")] = None,
    show_ids: Annotated[bool, typer.Option("--show-ids", "-I", help="Show internal IDs.")] = False,
):
    """Create a new experiment."""

    if all is True:
        include_expired = True
        include_archived = True

    experiments = ctx.obj.api.experiments_api.get_experiments(
        include_expired=include_expired,
        include_archived=include_archived,
    )

    columns = ["Name", "Description", "Created at", "Expires at"]
    if include_archived:
        columns.append("Archived")
    if show_ids:
        columns.insert(0, "ID")

    table = Table(*columns, title="Experiments")

    for exp in experiments.experiments:
        exp_expires_at = friendly_datetime(exp.expires_at)
        if exp.expires_at < datetime.now(UTC):
            exp_expires_at = f"[red]{exp_expires_at}"

        row = [
            exp.friendly_name,
            exp.description,
            friendly_datetime(exp.created_at),
            exp_expires_at,
        ]
        if include_archived:
            row.append("Yes" if exp.archived else "")
        if show_ids:
            row.insert(0, exp.id)

        table.add_row(*row, style="red" if exp.archived else "")

    console.print(table)


@experiment_app.command("show")
def show_experiment(
    ctx: Context,
    name: Annotated[str, typer.Argument(..., help="Name or ID of experiment to show.")],
):
    """Show experiment info."""
    if is_experiment_id(name):
        experiment = ctx.obj.api.experiments_api.get_experiment(experiment_id=name)
    else:
        experiments = ctx.obj.api.experiments_api.get_experiments(friendly_name=name)

        if not len(experiments.experiments) > 0:
            err_console.print(f"[red]Could not find experiment with name {name} :magnifying_glass_tilted_left:")
            raise typer.Exit(1)
        experiment = experiments.experiments[0]

    table = Table(box=None)
    table.add_column("Name", justify="right")
    table.add_column("Value")
    table.add_row("ID", experiment.id)
    table.add_row("Name", experiment.friendly_name)
    table.add_row("Created on", friendly_datetime(experiment.created_at))
    table.add_row("Expires on", friendly_datetime(experiment.expires_at))
    table.add_row("Deleted", "Yes" if experiment.deleted else "No")

    console.print(table)


@experiment_app.command("delete")
def delete_experiment(
    ctx: Context,
    name: Annotated[str, typer.Argument(..., help="Name or ID of experiment that must be deleted.")],
    force: Annotated[bool, typer.Option("--force", "-f", help="Do not ask for confirmation", is_flag=True)] = False,
):
    """Delete an experiment by name or ID."""
    if is_experiment_id(name):
        experiment_id = name
    else:
        experiments = ctx.obj.api.experiments_api.get_experiments(friendly_name=name)

        if not len(experiments.experiments) > 0:
            err_console.print(f"[red]Could not find experiment with name {name} :magnifying_glass_tilted_left:")
            raise typer.Exit(1)

        experiment_id = experiments.experiments[0].id

        if not force:
            typer.confirm(f"Are you sure you want to delete experiment {name} with ID '{experiment_id}'?", abort=True)

    try:
        deleted_experiment = ctx.obj.api.experiments_api.delete_experiment(experiment_id)
    except BadRequestException as err:
        err_console.print("[bold red]Could not delete experiment :x:")
        err_console.print(f"[red]{get_error_detail(err.body)}")
        raise typer.Exit(1) from err

    console.print(
        f":wastebasket: Deleted experiment '{deleted_experiment.friendly_name}' with ID '{deleted_experiment.id}'"
    )
